'From Squeak3.6 of ''6 October 2003'' [latest update: #5424] on 3 April 2004 at 10:07:57 pm'!
"Change Set:		RehashingSystemTracer-svp
Date:			3 April 2004
Author:			Stephen Pair

A subclass of SystemTracer2 that allows you to easily change object identityHashs.  See RehashingSystemTracer class comment.  SystemTracer2-ajh (3 April 2004) must be filed-in before this."!

SystemTracer2 subclass: #RehashingSystemTracer
	instanceVariableNames: 'lastHash hashMap '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tracing'!

!RehashingSystemTracer commentStamp: 'svp 8/23/2002 20:28' prior: 0!
I am just like SystemTracer2, except that I write new identity hashes for objects.  I rehash some special sets (like MethodDictionaries) prior to writing the clone image, and the clone image will rehash all objects when it starts up.

My instance variables:

lastHash - the lastHash that was assigned...I use the same algorithm to generate hashes as the VM
hashMap - A dictionary of all assigned identity hashes 

Subclasses can extend me if they need to write new images with modified hash values.!


!ProtoObject methodsFor: '*systemtracer2' stamp: 'svp 8/22/2002 21:45'!
rehashWithoutBecome
	"Do nothing.  Here so sending this to a Set does not have to do a time consuming respondsTo:"

	^self rehash! !


!RehashingSystemTracer methodsFor: 'rehashing' stamp: 'ajh 4/3/2004 19:04'!
initSwapMapPostShutdown
	"Subclasses may want to add converted objects to swap map before starting trace but after shuting down extra processes"

	| sets elementCount clone |
	sets _ self specialSets.
	elementCount _ sets inject: 0 into: [:c :set | c _ c + set size].
	hashMap := LargeIdentityDict new: elementCount.
	lastHash := 999.
	sets do: [ :ea |
		clone := ea cloneUsingHashesFrom: self.
		swapMap at: ea put: clone].
! !

!RehashingSystemTracer methodsFor: 'rehashing' stamp: 'ajh 4/3/2004 18:36'!
newHashFor: object
	"This is called by cloneUsingHashesFrom: (and compatibility with #hashMappedBy methods?)"

	^ hashMap at: object ifAbsentPut: [self newObjectHash]! !

!RehashingSystemTracer methodsFor: 'rehashing' stamp: 'ajh 4/3/2004 18:39'!
newObjectHash
	"Answer a new pseudo-random number for use as an identity hash."

	lastHash := 13849 + (27181 * lastHash) bitAnd: 65535.
	^ lastHash bitAnd: self hashMask
! !

!RehashingSystemTracer methodsFor: 'rehashing' stamp: 'ajh 4/3/2004 22:07'!
specialSets
	"The sets listed here get rehashed using the newly assigned identity hash values before the clone is written.  On startup, the clone image will rehash all objects to ensure that all hashed structures are properly arranged.  Rehashing is done in this two step process to retain maximum flexibility (if we tried to do everything up front, we would need to update the tracer every time a new type of Set is added with a different hashing and scanning algorithm, plus it would make the hashMap very large and slow)."

	^ MethodDictionary allInstances
! !

!RehashingSystemTracer methodsFor: 'clone startup' stamp: 'svp 8/22/2002 22:01'!
clonePreStartup

	Smalltalk rehashEntireImage! !

!RehashingSystemTracer methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 18:32'!
identityHashFor: object
	"Don not add to hashMap since this is only called once per object"

	^ hashMap at: object ifAbsent: [self newObjectHash]
! !

!RehashingSystemTracer methodsFor: 'object encoding' stamp: 'svp 8/21/2002 12:46'!
lastHash

	^lastHash! !


!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:24'!
compare: element with: object

	^element = object! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:48'!
findElementOrNil: anObject hash: hash
	"Answer the index of a first slot containing either a nil (indicating an empty slot) or an element that matches the given object. Answer the index of that slot or zero. Fail if neither a match nor an empty slot is found."

	| index |

	index _ self scanFor: anObject hash: hash.
	index > 0 ifTrue: [^index].

	"Bad scene.  Neither have we found a matching element
	nor even an empty slot.  No hashed set is ever supposed to get
	completely full."
	self error: 'There is no free space in this set!!'.! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:47'!
scanFor: anObject hash: hash
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	^self 
		scanFor: anObject 
		startingAt: (hash \\ array size) + 1
! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 10:09'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ array at: index) == nil or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array at: index) == nil or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !


!IdentityDictionary methodsFor: '*systemtracer2' stamp: 'ajh 4/3/2004 18:33'!
cloneUsingHashesFrom: tracer

	| object newDict |
	newDict := self shallowCopy.
	newDict withArray: (Array new: array size).
	1 to: array size do:
		[ :i |
		object := array at: i.
		object ifNotNil:
			[newDict
				noCheckAdd: object
				hash: (tracer newHashFor: object key)]].
	^newDict! !

!IdentityDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 18:36'!
noCheckAdd: anObject hash: hash
	"Variant to allow the hash to be specified
	by the caller.  8/23/02 svp"

	array at: (self findElementOrNil: anObject key hash: hash) put: anObject.
	tally _ tally + 1! !


!MethodDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:32'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ self basicAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ self basicAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!MethodDictionary methodsFor: '*systemtracer2' stamp: 'ajh 4/3/2004 18:33'!
cloneUsingHashesFrom: tracer

	| newDict key |
	newDict := self class basicNew: self basicSize.
	newDict init: array size.
	1 to: self basicSize do: 
		[ :i |
		key := self basicAt: i.
		key ifNotNil: 
			[newDict 
				noCheckAt: key 
				put: (array at: i) 
				hash: (tracer newHashFor: key)]].
	^newDict! !

!MethodDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:26'!
noCheckAt: key put: value hash: identityHash
	"Set the value at key to be value."
	| index |
	index _ self findElementOrNil: key hash: identityHash.
	(self basicAt: index) == nil
		ifTrue: 
			[tally _ tally + 1.
			self basicAt: index put: key].
	array at: index put: value.
	^ value! !


!String methodsFor: '*systemtracer2' stamp: 'svp 8/22/2002 22:45'!
hashMappedBy: map
	"My hash is independent of my oop."

	"While this statement is technically correct, it is misleading.  My hash does
	depend on the identity hash of my species, therefore it does need special care
	when cloning an image with new identity hashes - svp"

	^String stringHash: self initialHash: (self species hashMappedBy: map)! !


!Symbol methodsFor: '*systemtracer2' stamp: 'svp 11/15/2002 16:14'!
hashMappedBy: map
	"** DELETEME ** Overridden here to work with DVS packaging mechanism."

	^super hashMappedBy: map! !


!SystemDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 00:11'!
rehashEntireImage
	"Note, we cannot use allObjectsDo: because some objects like to swap identities
	during a rehash...this means that allObjectsDo: will not find all objects.  Also, this
	algorithm delays become operations to the very end where we can do them all at once. -svp"

	| objectCount allObjects numMdicts oldObjs newObjs newObject |

	numMdicts := MethodDictionary allInstances size.
	oldObjs := OrderedCollection new: numMdicts.
	newObjs := OrderedCollection new: numMdicts.

	"Create an array big enough to hold all objects in memory"
	objectCount := 0.
	self allObjectsDo: [ :ea | objectCount := objectCount + 1].
	allObjects := Array new: objectCount + 500.

	"Now fill the array"
	objectCount := 0.
	self allObjectsDo: [ :ea | allObjects at: (objectCount := objectCount + 1) put: ea ].

	'Rehashing the image: ', objectCount printString, ' objects' 
		displayProgressAt: Display center
		from: 1 to: objectCount
		during: [:bar | 
			objectCount := 0.
			allObjects do: [ :ea |
				bar value: (objectCount := objectCount + 1).
				newObject := ea rehashWithoutBecome.
				(newObject ~~ ea) ifTrue: 
					[oldObjs add: ea.
					newObjs add: newObject].
			].
			oldObjs asArray elementsExchangeIdentityWith: newObjs asArray.
		].! !


!WeakSet methodsFor: 'private' stamp: 'svp 8/23/2002 17:32'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements"

	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ array at: index) == flag or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array at: index) == flag or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!WeakSet methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:26'!
cloneUsingHashesFrom: tracer

	| newSet object |
	newSet := self class basicNew init: array size.
	1 to: array size do:
		[ :i |
		object := array at: i.
		(object == flag or: [object == nil]) ifFalse:
			[newSet 
				noCheckAdd: object
				hash: (object hashMappedBy: tracer)]].
	^newSet! !

!WeakSet methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:27'!
noCheckAdd: anObject hash: hash
	array at: (self findElementOrNil: anObject hash: hash) put: anObject.
	tally _ tally + 1! !

RehashingSystemTracer removeSelector: #hashMap!
RehashingSystemTracer removeSelector: #initialize:!

!RehashingSystemTracer reorganize!
('rehashing' initSwapMapPostShutdown newHashFor: newObjectHash specialSets)
('clone startup' clonePreStartup)
('object encoding' identityHashFor: lastHash)
!

